/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use buffer file
 * except in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.mmu.translator.coverage;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.mmu.basis.BufferAccessEvent;
import ru.ispras.microtesk.mmu.settings.BufferEventsSettings;
import ru.ispras.microtesk.mmu.test.sequence.engine.memory.MemoryAccess;
import ru.ispras.microtesk.mmu.test.sequence.engine.memory.MemoryAccessPath;
import ru.ispras.microtesk.mmu.test.sequence.engine.memory.MemoryAccessType;
import ru.ispras.microtesk.mmu.test.sequence.engine.memory.MemoryHazard;
import ru.ispras.microtesk.mmu.translator.ir.spec.MmuAddressType;
import ru.ispras.microtesk.mmu.translator.ir.spec.MmuBuffer;
import ru.ispras.microtesk.mmu.translator.ir.spec.MmuSegment;
import ru.ispras.microtesk.mmu.translator.ir.spec.MmuSubsystem;
import ru.ispras.microtesk.settings.AbstractSettings;
import ru.ispras.microtesk.settings.GeneratorSettings;
import ru.ispras.microtesk.settings.RegionSettings;

/**
 * @author <a href="mailto:kamkin@ispras.ru">Alexander Kamkin</a>
 */
public final class CoverageExtractor {
  private static final CoverageExtractor instance = new CoverageExtractor();

  public static CoverageExtractor get() {
    return instance;
  }

  private final Map<MmuAddressType, Collection<MemoryHazard>> addressHazards = new HashMap<>();
  private final Map<MmuBuffer, Collection<MemoryHazard>> bufferHazards = new HashMap<>();
  private final Map<MmuSubsystem, Collection<MemoryHazard>> memoryHazards = new HashMap<>();

  private final Map<MmuSubsystem, Map<MemoryAccessType, Collection<MemoryAccessPath>>> typeEnabledPaths =
      new HashMap<>();

  private final Map<MmuSubsystem, Map<MmuBuffer, Collection<MemoryAccessPath>>> bufferNormalPaths =
      new HashMap<>();

  private CoverageExtractor() {}

  public Collection<MemoryHazard> getHazards(final MmuAddressType address) {
    InvariantChecks.checkNotNull(address);

    Collection<MemoryHazard> coverage = addressHazards.get(address);
    if (coverage == null) {
      final AddressCoverageExtractor extractor = new AddressCoverageExtractor(address);
      addressHazards.put(address, coverage = extractor.getHazards());
    }

    return coverage;
  }

  public Collection<MemoryHazard> getHazards(final MmuBuffer device) {
    InvariantChecks.checkNotNull(device);

    Collection<MemoryHazard> coverage = bufferHazards.get(device);
    if (coverage == null) {
      final BufferCoverageExtractor extractor = new BufferCoverageExtractor(device);
      bufferHazards.put(device, coverage = extractor.getHazards());
    }

    return coverage;
  }

  private static boolean isEnabledPath(
      final MmuSubsystem memory, final MemoryAccessPath path, final GeneratorSettings settings) {
    InvariantChecks.checkNotNull(path);
    InvariantChecks.checkNotNull(settings);

    final Map<RegionSettings, Collection<MmuSegment>> segments =
        MemoryAccess.getPossibleSegments(path, settings);

    if (segments.isEmpty()) {
      return false;
    }

    final Collection<AbstractSettings> bufferEventsSettings = settings.get(BufferEventsSettings.TAG);
    for (final AbstractSettings section : bufferEventsSettings) {
      final BufferEventsSettings bufferEventsSection = (BufferEventsSettings) section;

      final MmuBuffer buffer = memory.getBuffer(bufferEventsSection.getName());
      InvariantChecks.checkNotNull(buffer);

      final Set<BufferAccessEvent> events = bufferEventsSection.getValues();
      InvariantChecks.checkNotNull(events);

      if (path.contains(buffer) && !events.contains(path.getEvent(buffer))) {
        return false;
      }
    }

    return true;
  }

  private static Collection<MemoryAccessPath> getEnabledPaths(
      final MmuSubsystem memory,
      final Collection<MemoryAccessPath> paths,
      final GeneratorSettings settings) {
    InvariantChecks.checkNotNull(memory);
    InvariantChecks.checkNotNull(paths);
    // Parameter {@code settings} can be null.

    if (settings == null) {
      return paths;
    }

    final Collection<MemoryAccessPath> enabledPaths = new ArrayList<>(paths.size());

    for (final MemoryAccessPath path : paths) {
      if (isEnabledPath(memory, path, settings)) {
        enabledPaths.add(path);
      }
    }

    return enabledPaths;
  }

  public Collection<MemoryAccessPath> getEnabledPaths(
      final MmuSubsystem memory, final MemoryAccessType type, final GeneratorSettings settings) {
    InvariantChecks.checkNotNull(memory);
    InvariantChecks.checkNotNull(type);
    // Parameter {@code settings} can be null.

    Map<MemoryAccessType, Collection<MemoryAccessPath>> typeToPaths = typeEnabledPaths.get(memory);
    if (typeToPaths == null) {
      typeEnabledPaths.put(memory,
          typeToPaths = new HashMap<MemoryAccessType, Collection<MemoryAccessPath>>());
    }

    Collection<MemoryAccessPath> paths = typeToPaths.get(type);
    if (paths == null) {
      final MemoryCoverageExtractor extractor = new MemoryCoverageExtractor(memory);
      typeToPaths.put(type, paths = extractor.getPaths(type));
    }

    return getEnabledPaths(memory, paths, settings);
  }

  public Collection<MemoryAccessPath> getNormalPaths(
      final MmuSubsystem memory, final MmuBuffer buffer, final GeneratorSettings settings) {
    InvariantChecks.checkNotNull(memory);
    InvariantChecks.checkNotNull(buffer);
    // Parameter {@code settings} can be null.

    Map<MmuBuffer, Collection<MemoryAccessPath>> bufferToPaths = bufferNormalPaths.get(memory);
    if (bufferToPaths == null) {
      bufferNormalPaths.put(memory,
          bufferToPaths = new HashMap<MmuBuffer, Collection<MemoryAccessPath>>());
    }

    Collection<MemoryAccessPath> paths = bufferToPaths.get(buffer);
    if (paths == null) {
      final MemoryCoverageExtractor extractor = new MemoryCoverageExtractor(memory);
      final Collection<MemoryAccessPath> allPaths = extractor.getPaths(null);
      final Collection<MemoryAccessPath> enabledPaths = getEnabledPaths(memory, allPaths, settings);

      paths = new ArrayList<>();
      for (final MemoryAccessPath path : enabledPaths) {
        if (path.contains(buffer) && path.contains(memory.getTargetBuffer())) {
          paths.add(path);
        }
      }

      bufferToPaths.put(buffer, paths);
    }

    return paths;
  }

  public Collection<MemoryHazard> getHazards(final MmuSubsystem memory) {
    InvariantChecks.checkNotNull(memory);

    Collection<MemoryHazard> coverage = memoryHazards.get(memory);
    if (coverage == null) {
      coverage = new ArrayList<>();
      for (final MmuBuffer device : memory.getBuffers()) {
        coverage.addAll(getHazards(device));
      }
      memoryHazards.put(memory, coverage);
    }

    return coverage;
  }
}
