/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru)
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.mmu.translator.generation.spec;

import static ru.ispras.fortress.util.InvariantChecks.checkNotNull;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.StandardOperation;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.fortress.util.Pair;
import ru.ispras.microtesk.basis.solver.integer.IntegerField;
import ru.ispras.microtesk.basis.solver.integer.IntegerVariable;
import ru.ispras.microtesk.mmu.translator.ir.AttributeRef;
import ru.ispras.microtesk.mmu.translator.ir.Buffer;
import ru.ispras.microtesk.mmu.translator.ir.Ir;
import ru.ispras.microtesk.mmu.translator.ir.Segment;

final class GuardPrinter {
  private final Ir ir;
  private final String context;

  private final Condition condTrue;
  private final Condition condFalse;

  public GuardPrinter(final Ir ir, final String context, final Node expr) {
    checkNotNull(ir);
    checkNotNull(context);
    checkNotNull(expr);

    this.ir = ir;
    this.context = context;

    this.condTrue = Condition.extract(expr);
    this.condFalse = condTrue.not();
  }

  public String getGuard() {
    return printGuard(condTrue);
  }

  public String getNegatedGuard() {
    return printGuard(condFalse);
  }

  public String getCondition() {
    return printCondition(condTrue);
  }

  private String printGuard(final Condition cond) {
    InvariantChecks.checkNotNull(cond);

    final List<Node> equalities = new ArrayList<>();
    final List<Node> segments = new ArrayList<>();
    final List<Node> buffers = new ArrayList<>();

    for (final Node atom : cond.getAtoms()) {
      if (isEquality(atom)) {
        equalities.add(atom);
      } else if (isSegmentAccess(atom)) {
        segments.add(atom);
      } else if (isBufferEvent(atom)) {
        buffers.add(atom);
      } else {
        throw new IllegalStateException("Illegal atomic condition: " + atom);
      }
    }

    InvariantChecks.checkFalse(
        equalities.isEmpty() && segments.isEmpty() && buffers.isEmpty());

    if (!equalities.isEmpty() && segments.isEmpty() && buffers.isEmpty()) {
      final List<String> conditionAtoms = extractEqualityConditionAtoms(equalities);
      InvariantChecks.checkNotEmpty(conditionAtoms);

      if (conditionAtoms.size() == 1) {
        return String.format("new MmuGuard(%s)", conditionAtoms.get(0));
      }

      final String operation =
          cond.getType() == Condition.Type.AND ? "and" : "or";

      return String.format("new MmuGuard(MmuCondition.%s(%s))",
          operation, Utils.toString(conditionAtoms, ", "));
    }

    if (equalities.isEmpty() && !segments.isEmpty() && buffers.isEmpty()) {
      InvariantChecks.checkTrue(segments.size() == 1, "One segment event is allowed.");

      final List<String> segmentIds = extractSegmentIds(segments.get(0));
      final String segmentIdsText = Utils.toString(segmentIds, ".get(), ") + ".get()";

      return String.format(
          "new MmuGuard(null, Arrays.<MmuSegment>asList(%s))", segmentIdsText);
    }

    if (equalities.isEmpty() && segments.isEmpty() && !buffers.isEmpty()) {
      InvariantChecks.checkTrue(buffers.size() == 1, "One buffer event is allowed.");
      final Pair<String, String> bufferEvent = extractBufferEventInfo(buffers.get(0));

      return String.format(
          "new MmuGuard(%s.get(), %s)", bufferEvent.first, bufferEvent.second);
    }

    throw new UnsupportedOperationException(
        "Guards for complex conditions that include several condition types " +
        "are not supported: " + cond);
  }

  private String printCondition(final Condition cond) {
    InvariantChecks.checkNotNull(cond);

    for (final Node atom : cond.getAtoms()) {
      if (!isEquality(atom)) {
        throw new IllegalStateException(
            "Only equality conditions are allowed: " + atom);
      } 
    }

    final List<String> conditionAtoms = extractEqualityConditionAtoms(cond.getAtoms());
    InvariantChecks.checkNotEmpty(conditionAtoms);

    final String operation =
        cond.getType() == Condition.Type.AND ? "and" : "or";

    return String.format("MmuCondition.%s(%s%s)",
        operation,
        System.lineSeparator() + "    ",
        Utils.toString(conditionAtoms, "," + System.lineSeparator() + "    "));
  }

  private static boolean isEquality(final Node node) {
    InvariantChecks.checkNotNull(node);

    if (node.getKind() == Node.Kind.OPERATION) {
      final NodeOperation op = (NodeOperation) node;
      return op.getOperationId() == StandardOperation.EQ ||
             op.getOperationId() == StandardOperation.NOTEQ;
    }

    return false;
  }

  private static boolean isSegmentAccess(final Node node) {
    InvariantChecks.checkNotNull(node);

    if (node.getKind() == Node.Kind.VARIABLE &&
        node.getUserData() instanceof AttributeRef) {
      final AttributeRef attrRef = (AttributeRef) node.getUserData();
      return attrRef.getTarget() instanceof Segment;
    }

    if (node.getKind() == Node.Kind.OPERATION) {
      final NodeOperation op = (NodeOperation) node;
      return op.getOperationId() == StandardOperation.NOT &&
             isSegmentAccess(op.getOperand(0));
    }

    return false;
  }

  private static boolean isBufferEvent(final Node node) {
    InvariantChecks.checkNotNull(node);

    if (node.getKind() == Node.Kind.VARIABLE &&
        node.getUserData() instanceof AttributeRef) {
      final AttributeRef attrRef = (AttributeRef) node.getUserData();
      return attrRef.getTarget() instanceof Buffer;
    }

    if (node.getKind() == Node.Kind.OPERATION) {
      final NodeOperation op = (NodeOperation) node;
      return op.getOperationId() == StandardOperation.NOT &&
             isBufferEvent(op.getOperand(0));
    }

    return false;
  }

  private List<String> extractSegmentIds(final Node node) {
    InvariantChecks.checkNotNull(node);

    if (node.getKind() == Node.Kind.VARIABLE) {
      final AttributeRef attrRef = (AttributeRef) node.getUserData();
      final Segment segment = (Segment) attrRef.getTarget();
      return Collections.singletonList(segment.getId());
    }

    if (node.getKind() == Node.Kind.OPERATION) {
      final NodeOperation op = (NodeOperation) node;
      InvariantChecks.checkTrue(op.getOperationId() == StandardOperation.NOT);

      final AttributeRef attrRef = (AttributeRef) op.getOperand(0).getUserData();
      final Segment segment = (Segment) attrRef.getTarget();

      final Set<String> allSegmentIds = new HashSet<>(ir.getSegments().keySet());
      allSegmentIds.remove(segment.getId());

      return new ArrayList<>(allSegmentIds);
    }

    throw new IllegalArgumentException("Illegal segment access expression: " + node);
  }

  private Pair<String, String> extractBufferEventInfo(final Node node) {
    InvariantChecks.checkNotNull(node);

    if (node.getKind() == Node.Kind.VARIABLE) {
      final AttributeRef attrRef = (AttributeRef) node.getUserData();
      final Buffer buffer = (Buffer) attrRef.getTarget();
      return new Pair<>(buffer.getId(), "BufferAccessEvent.HIT");
    }

    if (node.getKind() == Node.Kind.OPERATION) {
      final NodeOperation op = (NodeOperation) node;
      InvariantChecks.checkTrue(op.getOperationId() == StandardOperation.NOT);

      final AttributeRef attrRef = (AttributeRef) op.getOperand(0).getUserData();
      final Buffer buffer = (Buffer) attrRef.getTarget();
      return new Pair<>(buffer.getId(), "BufferAccessEvent.MISS");
    }

    throw new IllegalArgumentException("Illegal buffer event expression: " + node);
  }

  private List<String> extractEqualityConditionAtoms(final List<Node> nodes) {
    InvariantChecks.checkNotEmpty(nodes);

    final List<String> result = new ArrayList<>();
    for (final Node node : nodes) {
      final NodeOperation op = (NodeOperation) node;
      InvariantChecks.checkTrue(
          op.getOperationId() == StandardOperation.EQ ||
          op.getOperationId() == StandardOperation.NOTEQ);

      final Atom lhs = AtomExtractor.extract(op.getOperand(0));
      final Atom rhs = AtomExtractor.extract(op.getOperand(1));

      final BigInteger value;
      final Atom variableAtom;

      if (Atom.Kind.VALUE == lhs.getKind() && Atom.Kind.VALUE == rhs.getKind()) {
        throw new IllegalArgumentException(String.format(
            "Both sides of an equality expression are constants: %s = %s",
            op.getOperand(0), op.getOperand(1)));
      } else if (Atom.Kind.VALUE == lhs.getKind()) {
        value = (BigInteger) lhs.getObject();
        variableAtom = rhs;
      } else if (Atom.Kind.VALUE == rhs.getKind()) {
        value = (BigInteger) rhs.getObject();
        variableAtom = lhs;
      } else {
        throw new IllegalArgumentException(String.format( 
            "Both sides of an equality expression are variables: %s = %s",
            op.getOperand(0), op.getOperand(1)));
      }

      final String variableText;
      final String valueText = Utils.toString(value);

      switch (variableAtom.getKind()) {
        case VARIABLE: {
          final IntegerVariable intVar = (IntegerVariable) variableAtom.getObject();
          variableText = Utils.getVariableName(context, intVar.getName());
          break;
        }

        case FIELD: {
          final IntegerField intField = (IntegerField) variableAtom.getObject();
          variableText = Utils.toString(context, intField);
          break;
        }

        case CONCAT: {
          @SuppressWarnings("unchecked")
          final List<IntegerField> fields = (List<IntegerField>) variableAtom.getObject();
          final List<String> fieldTexts = new ArrayList<>();

          for (final IntegerField field : fields) {
            fieldTexts.add(Utils.toString(context, field, false));
          }

          variableText = String.format(
            "MmuExpression.rcat(%s)", Utils.toString(fieldTexts, ", "));
          break;
        }

        default: {
          throw new IllegalArgumentException("Variable is expected.");
        }
      }

      final String operationText =
          op.getOperationId() == StandardOperation.EQ ? "eq" : "neq";

      final String conditionAtom = String.format("MmuConditionAtom.%s(%s, %s)",
          operationText, variableText, valueText);

      result.add(conditionAtom);
    }

    return result;
  }
}
